"""jello - query JSON at the command line with python syntax"""


__version__ = '1.6.1'
AUTHOR = 'Kelly Brazil'
WEBSITE = 'https://github.com/kellyjonbrazil/jello'
COPYRIGHT = '© 2020-2025 Kelly Brazil'
LICENSE = 'MIT License'
